import 'dart:convert';

import 'package:dio/dio.dart';
import 'package:flutter/foundation.dart';
import 'package:flutter_easyloading/flutter_easyloading.dart';
import 'package:get/get.dart' as pp;
import '/url.dart';
import 'package:pretty_dio_logger/pretty_dio_logger.dart';

///Global network request `dio` instance singleton `XHttp`
///Store api endpoint URL in `lib/url.dart'`
///like :
///{@tool snippet}
///```dart
///class Urls {
///  static const BASE_URL = 'localhost/api';
///}
///```
///{@end-tool}
///A simple get request is like
//////{@tool snippet}
///```dart
///  void getRequest(){
///     XHttp.request(Urls.BASE_URL, method: XHttp.GET);
///   }
///```
///{@end-tool}
///
///
class XHttp {
  static const String GET = "GET";
  static const String POST = "POST";
  static const String PUT = "PUT";
  static const String PATCH = "PATCH";
  static const String DELETE = "DELETE";

  static const CUSTOM_ERROR_CODE = 'DIO_CUSTOM_ERROR';
  static const REQUEST_TYPE_STR = 'REQUEST';
  static const RESPONSE_TYPE_STR = 'RESPONSE';
  static const ERROR_TYPE_STR = 'RESPONSE_ERROR';
  static const DEFAULT_LOAD_MSG = 'Please wait...';

  static const CONNECT_TIMEOUT = 60000;
  static const RECEIVE_TIMEOUT = 60000;
  static const SEND_TIMEOUT = 60000;

  static const DIALOG_TYPE_OTHERS = 'OTHERS';
  static const DIALOG_TYPE_TOAST = 'TOAST';
  static const DIALOG_TYPE_ALERT = 'ALERT';
  static const DIALOG_TYPE_CUSTOM = 'CUSTOM';

  static String loadMsg = DEFAULT_LOAD_MSG;

  static String errorShowTitle = 'Something went wrong';

  static String? errorShowMsg;

  static CancelToken cancelToken = CancelToken();

  static CancelToken whiteListCancelToken = CancelToken();

  static Dio? dio;

  String _getBaseUrl() {
    return Urls.BASE_URL;
  }

  XHttp._internal() {
    if (null == dio) {
      dio = Dio(BaseOptions(
        baseUrl: _getBaseUrl(),
        // contentType: ,
        // responseType: ,
        responseType: ResponseType.plain,
        connectTimeout: CONNECT_TIMEOUT,
        receiveTimeout: RECEIVE_TIMEOUT,
        sendTimeout: SEND_TIMEOUT,
      ));
      _init();
    }
  }

  static final XHttp _instance = XHttp._internal();

  void _init() {
    dio?.interceptors.add(
      InterceptorsWrapper(
        onRequest: (RequestOptions options, handler) async {
          if (kDebugMode) {
            print("before request");
          }
          _handleRequest(options, handler);
          // 有 token 时，添加 token。放打印日志后面，避免泄露 token。
          // 也可以登录成功后掉用 XHttp.setToken() 方法设置 token，但是持久化的话还是要这样最好。
          // if (Storage.box.get(Constants.isUserLoggedIn, defaultValue: false)) {
          //   String? token =
          //       Storage.box.get(Constants.authToken, defaultValue: null);
          //   // if (token != dio?.options.headers['authorization']) {
          //   //   dio?.options.headers['authorization'] = token;
          //   //   options.headers['authorization'] = token; // 不设置的话第一次的请求会有问题，上面的是全局设置尚未对本条请求生效。
          //   // }
          //   // options.headers["Content-Type"] = "application/json";
          //   // dio?.options.headers['Content-Type'] = "application/json";
          //   if (token != null) {
          //     options.headers["API-Auth"] = token;

          //     // options.headers["DeviceId"] = Storage.box.get(Constants.deviceId, defaultValue: '');
          //     dio?.options.headers['API-Auth'] = token;

          //     // dio?.options.headers['DeviceId'] = Storage.box.get(Constants.deviceId, defaultValue: '');
          //   }

          //   Logger().e(
          //       "DeviceId => ${Storage.box.get(Constants.deviceId, defaultValue: '')}");
          //   Logger().e("token => $token");
          // }
          options.headers["API-KEY"] =
              pp.GetPlatform.isAndroid ? '12345' : '12345';
          dio?.options.headers['API-KEY'] =
              pp.GetPlatform.isAndroid ? '12345' : '12345';

          return handler.next(options);
        },
        onResponse: (Response response, ResponseInterceptorHandler handler) {
          if (kDebugMode) {
            print("before responding");
          }
          // _handleResponse(response, handler);
          RequestOptions option = response.requestOptions;
          int? code;
          code = jsonDecode(response.data)["statusCode"];
          // print('nowfal $code');
          // String code = (response.data ?? {})['statusCode'];
          String msg = "${response.statusMessage}";
          // String msg = (response.data ?? {})['msg'] ?? response.statusMessage;
          // Static data or based on the actual return structure analysis in the background, that is, when code == '0', data is valid data.
          bool isSuccess = code == 200;
          // bool isSuccess = option.contentType != null && option.contentType!.contains("text");
          _handleResponse(response, handler, isSuccess);
          response.data = Result(
              response.data, isSuccess, code ?? response.statusCode!, msg,
              headers: response.headers);
          return handler.next(response);
        },
        onError: (DioError error, handler) {
          if (kDebugMode) {
            print("before error");
          }
          _handleError(error);

          if (error.response != null && error.response?.data != null) {
            error.response!.data = Result(
                error.response!.data,
                false,
                error.response!.statusCode!,
                errorShowMsg ?? error.response!.statusMessage!,
                headers: error.response?.headers);
          } else {
            throw Exception(errorShowMsg);
          }
          return handler.next(error);
        },
      ),
    );
    dio?.interceptors.add(PrettyDioLogger(
      requestHeader: true,
      requestBody: true,
      responseBody: true,
      responseHeader: false,
      compact: false,
    ));
  }

  void _handleRequest(RequestOptions options, handler) {
    Toast.hide();
    Toast.loading(loadMsg);
    Map logData = {
      'url': options.baseUrl + options.path,
      'method': options.method,
      'headers': options.headers,
      'data': options.data ?? options.queryParameters,
    };
    _dealRequestInfo(logData, REQUEST_TYPE_STR);
  }

  void _handleResponse(Response response, handler, bool isSuccess) {
    Map logData = {
      'url': response.requestOptions.uri,
      'method': response.requestOptions.method,
      'headers': response.headers,
      'data': response.data,
      'statusCode': response.statusCode,
      'statusMessage': response.statusMessage,
    };
    _dealRequestInfo(logData, RESPONSE_TYPE_STR);
    if (!isSuccess) {
      Toast.show("Something went wrong..", type: Toast.ERROR);
    } else {
      Toast.hide();
    }
  }

  void _handleError(DioError error) {
    String errorTypeInfo = 'other errors！';
    switch (error.type) {
      case DioErrorType.connectTimeout:
        errorTypeInfo = 'Connection timed out！';
        break;
      case DioErrorType.sendTimeout:
        errorTypeInfo = "Request timed out！";
        break;
      case DioErrorType.receiveTimeout:
        errorTypeInfo = "Response timed out！";
        break;
      case DioErrorType.response:
        errorTypeInfo = "Service exception！";
        break;
      case DioErrorType.cancel:
        errorTypeInfo = "request cancellation！";
        break;
      case DioErrorType.other:
        errorTypeInfo = "${error.message}";
        break;
      default:
        break;
    }
    Map logData = {
      'url': error.requestOptions.baseUrl + error.requestOptions.path,
      'method': error.requestOptions.method,
      'headers': error.response?.headers,
      'data': error.response?.data,
      'statusCode': error.response?.statusCode,
      'statusMessage': error.response?.statusMessage,
      'errorType': error.type,
      'errorMessage': error.message,
      'errorTypeInfo': errorTypeInfo,
    };
    _dealRequestInfo(logData, ERROR_TYPE_STR);
    Toast.hide();
    errorShowMsg =
        "$errorShowTitle ${error.response?.statusCode ?? 'unknown'} $errorTypeInfo \n ${error.response?.statusMessage ?? ''} \n ${error.response?.data ?? ''}";
  }

  String _dealRequestInfo(Map logData, String logType) {
    String logStr = "\n";
    logStr +=
        "========================= $logType START =========================\n";
    logStr += "- URL: ${logData['url']} \n";
    logStr += "- METHOD: ${logData['method']} \n";
    // logStr += "- HEADER: \n { \n";
    // logStr += parseData(logData['headers']);
    // logStr += "\n } \n";
    if (logData['data'] != null) {
      logStr += "- ${logType}_BODY: \n";
      logStr +=
          "!!!!!----------*!*##~##~##~##*!*##~##~##~##*!*----------!!!!! \n";
      logStr += "${parseData(logData['data'])} \n";
      logStr +=
          "!!!!!----------*!*##~##~##~##*!*##~##~##~##*!*----------!!!!! \n";
    }
    if (logType.contains(RESPONSE_TYPE_STR)) {
      logStr += "- STATUS_CODE: ${logData['statusCode']} \n";
      logStr += "- STATUS_MSG: ${logData['statusMessage']} \n";
    }
    if (logType == ERROR_TYPE_STR) {
      logStr += "- ERROR_TYPE: ${logData['errorType']} \n";
      logStr += "- ERROR_MSG: ${logData['errorMessage']} \n";
      logStr += "- ERROR_TYPE_INFO: ${logData['errorTypeInfo']} \n";
    }
    logStr +=
        "========================= $logType E N D =========================\n";
    logWrapped(logStr);
    return logStr;
  }

  Future _showResultDialog(Response? response, resultDialogConfig) async {
    if (response == null) {
      return;
    }
    resultDialogConfig = resultDialogConfig ?? {};
    String dialogType = resultDialogConfig['type'] ?? XHttp.DIALOG_TYPE_TOAST;
    if (dialogType == XHttp.DIALOG_TYPE_OTHERS) {
      return;
    }
    bool isSuccess = response.data?.success ?? false;
    String msg = response.data?.msg ?? 'unknown mistake';
    if (dialogType == XHttp.DIALOG_TYPE_TOAST) {
      isSuccess
          ? Toast.show(resultDialogConfig['successMsg'] ?? msg,
              type: Toast.SUCCESS)
          : Toast.show(resultDialogConfig['errorMsg'] ?? msg,
              type: Toast.ERROR);
      return;
    }
    if (dialogType == XHttp.DIALOG_TYPE_ALERT) {
      return;
    }
    if (dialogType == XHttp.DIALOG_TYPE_CUSTOM) {
      if (isSuccess) {
        if (resultDialogConfig['onSuccess'] != null) {
          resultDialogConfig['onSuccess'](response.data);
        }
      } else {
        if (resultDialogConfig['onError'] != null) {
          resultDialogConfig['onError'](response.data);
        }
      }
    }
  }

  void _catchOthersError(e) {
    String errMsg =
        "${errorShowMsg ?? e}$CUSTOM_ERROR_CODE".split(CUSTOM_ERROR_CODE)[0];
    int errMsgLength = errMsg.length;
    String errshowMsg = errMsgLength > 300 ? errMsg.substring(0, 150) : errMsg;
    if (e is DioError) {
      if (CancelToken.isCancel(e)) {
        Toast.show('Cancel Request Successful');
        return;
      }
      Toast.show(errshowMsg, type: Toast.WARNING);
      return;
    }
    Toast.show(errshowMsg + "\n......", type: Toast.ERROR);
  }

  static XHttp getInstance({String? baseUrl, String? msg}) {
    String targetBaseUrl = baseUrl ?? _instance._getBaseUrl();
    loadMsg = msg ?? DEFAULT_LOAD_MSG;
    if (dio?.options.baseUrl != targetBaseUrl) {
      dio?.options.baseUrl = targetBaseUrl;
    }
    return _instance;
  }

  static XHttp cancelRequest() {
    Toast.hide();
    cancelToken.cancel('cancel request');
    cancelToken = CancelToken();
    return _instance;
  }

  static XHttp cancelWhiteListRequest() {
    Toast.hide();
    whiteListCancelToken.cancel('cancel whiteList request');
    whiteListCancelToken = CancelToken();
    return _instance;
  }

  static CancelToken getCancelToken() {
    return cancelToken;
  }

  static CancelToken getWhiteListCancelToken() {
    return whiteListCancelToken;
  }

  static CancelToken getNewCancelToken() {
    return CancelToken();
  }

  Future get(String url,
      [Map<String, dynamic>? params,
      resultDialogConfig,
      bool isCancelWhiteList = false]) async {
    Response? response;
    CancelToken setCancelToken =
        isCancelWhiteList ? whiteListCancelToken : cancelToken;
    try {
      if (params != null) {
        response = await dio!
            .get(url, queryParameters: params, cancelToken: setCancelToken);
        return response.data;
      } else {
        response = await dio!.get(url, cancelToken: setCancelToken);
        return response.data;
      }
    } catch (e) {
      _catchOthersError(e);
    } finally {
      _showResultDialog(response!, resultDialogConfig);
    }
  }

  Future post(String url,
      [Map<String, dynamic>? data,
      resultDialogConfig,
      bool isCancelWhiteList = false]) async {
    Response? response;
    try {
      response = await dio?.post(url,
          data: data,
          cancelToken: isCancelWhiteList ? whiteListCancelToken : cancelToken);
      return response?.data;
    } catch (e) {
      _catchOthersError(e);
    } finally {
      _showResultDialog(response, resultDialogConfig);
    }
  }

  Future put(String url,
      [Map<String, dynamic>? data,
      resultDialogConfig,
      bool isCancelWhiteList = false]) async {
    Response? response;
    try {
      response = await dio?.put(url,
          data: data,
          cancelToken: isCancelWhiteList ? whiteListCancelToken : cancelToken);
      return response?.data;
    } catch (e) {
      _catchOthersError(e);
    } finally {
      _showResultDialog(response, resultDialogConfig);
    }
  }

  Future patch(String url,
      [Map<String, dynamic>? data,
      resultDialogConfig,
      bool isCancelWhiteList = false]) async {
    Response? response;
    try {
      response = await dio?.patch(url,
          data: data,
          cancelToken: isCancelWhiteList ? whiteListCancelToken : cancelToken);
      return response?.data;
    } catch (e) {
      _catchOthersError(e);
    } finally {
      _showResultDialog(response, resultDialogConfig);
    }
  }

  Future delete(String url,
      [Map<String, dynamic>? data,
      resultDialogConfig,
      bool isCancelWhiteList = false]) async {
    Response? response;
    try {
      response = await dio?.delete(url,
          data: data,
          cancelToken: isCancelWhiteList ? whiteListCancelToken : cancelToken);
      return response?.data;
    } catch (e) {
      _catchOthersError(e);
    } finally {
      _showResultDialog(response, resultDialogConfig);
    }
  }

  static Future request(
    String url, {
    String method = XHttp.GET,
    Map<String, dynamic>? queryParameters,
    Map<String, dynamic>? data,
    bool isCancelWhiteList = false,
    resultDialogConfig,
    Options? options,
    void Function(int, int)? onSendProgress,
    void Function(int, int)? onReceiveProgress,
    String? msg,
    String? baseUrl,
  }) async {
    XHttp.getInstance(baseUrl: baseUrl, msg: msg);
    Response? response;
    try {
      response = await dio?.request(
        url,
        options: options ??
            Options(
                method: method, contentType: Headers.formUrlEncodedContentType),
        queryParameters: queryParameters,
        data: data,
        cancelToken: isCancelWhiteList ? whiteListCancelToken : cancelToken,
        onReceiveProgress: onReceiveProgress,
        onSendProgress: onSendProgress,
      );
      return response?.data;
    } catch (e) {
      _instance._catchOthersError(e);
    } finally {
      _instance._showResultDialog(
        response,
        resultDialogConfig ?? {'type': XHttp.DIALOG_TYPE_OTHERS},
      );
    }
  }

  Future downloadFile(urlPath, savePath,
      [resultDialogConfig, bool isCancelWhiteList = false]) async {
    Response? response;
    try {
      response = await dio?.download(urlPath, savePath,
          onReceiveProgress: (int count, int total) {
        print("$count $total");
      }, cancelToken: isCancelWhiteList ? whiteListCancelToken : cancelToken);
      return response?.data;
    } catch (e) {
      _catchOthersError(e);
    } finally {
      _showResultDialog(response, resultDialogConfig);
    }
  }

  // /// post  【Web】
  // Future postForm(String url, [Map<String, dynamic> params, resultDialogConfig, bool isCancelWhiteList = false]) async {
  //   Response response;
  //   try {
  //     response = await dio.post(url,
  //         queryParameters: params, cancelToken: isCancelWhiteList ? whiteListCancelToken : cancelToken);
  //     return response.data;
  //   } catch (e) {
  //     _catchOthersError(e);
  //   } finally {
  //     _showResultDialog(response, resultDialogConfig);
  //   }
  // }

  static String? getBaseUrl() {
    return dio?.options.baseUrl;
  }

  static XHttp setBaseUrl(String baseUrl) {
    dio?.options.baseUrl = baseUrl;
    return _instance;
  }

  static Map? getHeaders() {
    return dio?.options.headers;
  }

  static dynamic getHeader(String key) {
    return dio?.options.headers[key];
  }

  static XHttp setHeaders(Map headers) {
    dio?.options.headers = headers.cast<String, dynamic>();
    return _instance;
  }

  static XHttp setHeader(String key, String value) {
    dio?.options.headers[key] = value;
    return _instance;
  }

  static XHttp removeHeader(String key) {
    dio?.options.headers.remove(key);
    return _instance;
  }

  static XHttp removeAllHeaders() {
    dio?.options.headers.clear();
    return _instance;
  }

  static Map getRequestTimeout() {
    return {
      'connectTimeout': dio?.options.connectTimeout,
      'receiveTimeout': dio?.options.receiveTimeout,
      'sendTimeout': dio?.options.sendTimeout
    };
  }

  static XHttp setRequestTimeout(int timeout) {
    dio?.options.connectTimeout = timeout;
    dio?.options.receiveTimeout = timeout;
    dio?.options.sendTimeout = timeout;
    return _instance;
  }

  static XHttp setConnectTimeout(int timeout) {
    dio?.options.connectTimeout = timeout;
    return _instance;
  }

  static XHttp setReceiveTimeout(int timeout) {
    dio?.options.receiveTimeout = timeout;
    return _instance;
  }

  static XHttp setSendTimeout(int timeout) {
    dio?.options.sendTimeout = timeout;
    return _instance;
  }

  static Map<String, dynamic>? getAuthUser() {
    String? token = dio?.options.headers['authorization'];
    if (null == token) {
      return null;
    }

    return {'account': 'xxx', 'name': 'xxx', 'roles': 'xxx'};
  }

  static XHttp setAuthToken([String? token]) {
    // if (null == token) {
    //   dio?.options.headers.remove('API-Auth');
    //   dio?.options.headers.remove('API-KEY');
    // } else {
    //   dio?.options.headers['API-Auth'] = token;
    //   dio?.options.headers['API-KEY'] = '12345';
    // }
    return _instance;
  }

  static XHttp setErrorTitle(String msg) {
    errorShowTitle = msg;
    return _instance;
  }

// static XHttp setContentType(String contentType) {
//   dio.options.contentType = contentType;
//   return _instance;
// }

// static XHttp setContentTypeMultipartForm() {
//   dio.options.contentType = "multipart/form-data";
//   return _instance;
// }

// static XHttp setDataType(ResponseType dataType) {
//   dio.options.responseType = dataType;
//   return _instance;
// }

// static XHttp setDataTypeJson() {
//   dio.options.responseType = ResponseType.json;
//   return _instance;
// }

}

String parseData(data) {
  String responseStr = "";
  if (data is Map) {
    responseStr += data.mapToStructureString();
  } else if (data is FormData) {
    final formDataMap = Map()
      ..addEntries(data.fields)
      ..addEntries(data.files);
    responseStr += formDataMap.mapToStructureString();
  } else if (data is List) {
    responseStr += data.listToStructureString();
  } else {
    responseStr += data.toString();
  }
  return responseStr;
}

void logWrapped(String text) {
  final pattern = RegExp('.{1,800}'); // 800 is the size of each chunk
  pattern.allMatches(text).forEach((match) => print(match.group(0)));
}

extension Map2StringEx on Map {
  String mapToStructureString({int indentation = 0, String space = "  "}) {
    if (this == null || this.isEmpty) {
      return "$this";
    }
    String result = "";
    String indentationContent = space * indentation;
    result += "{";
    this.forEach((key, value) {
      if (value is Map) {
        result += "\n$indentationContent" +
            "\"$key\": ${value.mapToStructureString(indentation: indentation + 1)},";
      } else if (value is List) {
        result += "\n$indentationContent" +
            "\"$key\": ${value.listToStructureString(indentation: indentation + 1)},";
      } else {
        result += "\n$indentationContent" +
            "\"$key\": ${value is String ? "\"$value\"," : "$value,"}";
      }
    });
    result = result.substring(0, result.length - 1);
    result += "\n$indentationContent}";
    return result;
  }
}

extension List2StringEx on List {
  String listToStructureString({int indentation = 0, String space = "  "}) {
    if (this == null || this.isEmpty) {
      return "$this";
    }
    String result = "";
    String indentationContent = space * indentation;
    result += "[";
    for (var value in this) {
      if (value is Map) {
        result += "\n$indentationContent" +
            space +
            "${value.mapToStructureString(indentation: indentation + 1)},"; // 加空格更好看
      } else if (value is List) {
        result += value.listToStructureString(indentation: indentation + 1);
      } else {
        result += "\n$indentationContent" + value is String
            ? "\"$value\","
            : "$value,";
      }
    }
    result = result.substring(0, result.length - 1);
    result += "\n$indentationContent]";
    return result;
  }
}

class Result {
  var data;
  bool success;
  int code;
  String msg;
  var headers;
  Result(this.data, this.success, this.code, this.msg, {this.headers});
}

class Toast {
  Toast._() {}
  static final Toast _instance = Toast._();

  static const String SUCCESS = "SUCCESS";
  static const String ERROR = "ERROR";
  static const String WARNING = "WARNING";
  static const String INFO = "INFO";

  static loading(String msg) {
    EasyLoading.show(status: msg);
  }

  static progeress(double value, String msg) {
    EasyLoading.showProgress(value, status: msg);
  }

  static show(String msg, {String? type}) {
    switch (type) {
      case Toast.SUCCESS:
        EasyLoading.showSuccess(msg);
        break;
      case Toast.ERROR:
        EasyLoading.showError(msg);
        break;
      case Toast.WARNING:
        EasyLoading.showInfo(msg);
        break;
      case Toast.INFO:
      default:
        EasyLoading.showToast(msg);
        break;
    }
  }

  static hide() {
    EasyLoading.dismiss();
  }
}

//  XHttp.getInstance().post("/user/login", {
//     "username": username,
//     "password": password
//   }).then((res) {
//     // DO SOMETHING
//   }).catchError((err) {
//     // DO SOMETHING
//   });
